/* logger.c
 *
 * Simon Melhuish
 * Cardiff Astronomy Instrumentation Group
 * 2004
 *
 * Logger functions for owwlog
 */
 
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

//#ifdef _WIN32
//#  include "win32config.h"
//#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "owwl.h"
#include "logger.h"
 
/* Messages, warnings and errors */
void owwlog_message(char* format, ...) ;
void owwlog_warning(char* format, ...) ;
void owwlog_error(char* format, ...) ;

static int 
logger_accumulate_floats(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerFloats *logF ;
  int i ;
  
  log = (Logger *) data->user_data ;
  
  logF = (LoggerFloats *) log->internal ;
  
  for (i=0; i<logF->count; ++i)
  {
    logF->sum[i] += (double) data->device_data.floats.floats[i] ;
  }
  
  ++log->N ;
  
  return LOGGER_OK ;
} 

static int 
logger_accumulate_wind(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerWind *logW ;
  double rad ;
  
  log = (Logger *) data->user_data ;
  
  logW = (LoggerWind *) log->internal ;
  
  logW->s_sum += data->device_data.wind.speed ;
  if (data->device_data.wind.gust > logW->s_max)
    logW->s_max = data->device_data.wind.gust ;
  
  rad = data->device_data.wind.bearing * 0.017453292 ;
  
  logW->x_sum += data->device_data.wind.speed * cos(rad) ;
  logW->y_sum += data->device_data.wind.speed * sin(rad) ;
      
  ++log->N ;
  
  return LOGGER_OK ;
} 

static int 
logger_accumulate_rain(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerRain *logR ;
  
  log = (Logger *) data->user_data ;
  
  logR = (LoggerRain *) log->internal ;
  
  logR->r_sum += data->device_data.rain.rain_rate ;
  logR->r = data->device_data.rain.rain_since_reset ;
  logR->t = data->device_data.rain.rain_reset_time ;
  logR->c = data->device_data.rain.rain_count ;
  
  ++log->N ;
  
  return LOGGER_OK ;
} 

static int 
logger_accumulate_gpc(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerGpc *logG ;
  
  log = (Logger *) data->user_data ;
  
  logG = (LoggerGpc *) log->internal ;
  
  logG->r_sum += data->device_data.gpc.gpc_rate ;
  logG->r = data->device_data.gpc.gpc_calibrated ;
  logG->t = data->device_data.gpc.gpc_reset_time ;
  logG->c = data->device_data.gpc.gpc_count ;
  logG->d += data->device_data.gpc.gpc_delta ;
  
  ++log->N ;
  
  return LOGGER_OK ;
} 

static int 
logger_reset_floats(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerFloats *logF ;
  int i ;
    
  log = (Logger *) data->user_data ;
  
  logF = (LoggerFloats *) log->internal ;
  
  for (i=0; i<logF->count; ++i)
  {
    logF->sum[i] = 0.0 ;
  }
  
  return LOGGER_OK ;
}

static int 
logger_reset_wind(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerWind *logW ;
    
  log = (Logger *) data->user_data ;
  
  logW = (LoggerWind *) log->internal ;
  
  logW->s_sum = logW->x_sum = logW->y_sum = logW->s_max = 0.0 ;
  
  return LOGGER_OK ;
}

static int 
logger_reset_rain(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerRain *logR ;
    
  log = (Logger *) data->user_data ;
  
  logR = (LoggerRain *) log->internal ;
  
  logR->c = 0L ;
  logR->r_sum = logR->r = 0.0 ;
  logR->t = 0 ;
  
  return LOGGER_OK ;
}

static int 
logger_reset_gpc(owwl_data *data, void *user_data)
{
  Logger *log ;
  LoggerGpc *logG ;
    
  log = (Logger *) data->user_data ;
  
  logG = (LoggerGpc *) log->internal ;
  
  logG->c = logG->d = 0L ;
  logG->r_sum = logG->r = 0.0 ;
  logG->t = 0 ;
  
  return LOGGER_OK ;
}

static int 
logger_result_floats(owwl_data *data, void *user_data, double *result, int arg)
{
  Logger *log ;
  LoggerFloats *logF ;
    
  log = (Logger *) data->user_data ;
  
  logF = (LoggerFloats *) log->internal ;
  
  if ((arg >= 0) && (arg < logF->count))
  {
    *result = logF->sum[arg] / (double) log->N ;
    return LOGGER_OK ;
  }
  
  return LOGGER_ERROR ;    
}

static int 
logger_result_wind(owwl_data *data, void *user_data, double *result, int arg)
{
  Logger *log ;
  LoggerWind *logW ;
    
  log = (Logger *) data->user_data ;
  
  logW = (LoggerWind *) log->internal ;
  
  switch (arg)
  {
    case 0: /* Speed */
      *result = logW->s_sum / (double) log->N ;
      return LOGGER_OK ;
      
    case 1: /* Gust */
      *result = logW->s_max ;
      return LOGGER_OK ;
      
    case 2: /* Bearing */
    {
      double a ;
      a = 57.29577951 * atan2(logW->y_sum, logW->x_sum) ;
      *result = (a < 0.0) ? -a : 360.0 - a ;
      return LOGGER_OK ;
    }   
  }
  
  return LOGGER_ERROR ;    
}

static int 
logger_result_rain(owwl_data *data, void *user_data, double *result, int arg)
{
  Logger *log ;
  LoggerRain *logR ;
    
  log = (Logger *) data->user_data ;
  
  logR = (LoggerRain *) log->internal ;
  
  switch (arg)
  {
    case 0: /* Rain since reset */
      *result = logR->r ;
      return LOGGER_OK ;
    break ;
      
    case 1: /* Reset date / time */
      *result = (double) logR->t ;
      return LOGGER_OK ;
    break ;
      
    case 2: /* Rain rate */
      *result = logR->r_sum / (double) log->N ;
      return LOGGER_OK ;
    break ;
     
    case 3: /* Monotonic count */
      *result = (double) logR->c ;
      return LOGGER_OK ;
    break ;
  }
  
  return LOGGER_ERROR ;    
}

static int 
logger_result_gpc(owwl_data *data, void *user_data, double *result, int arg)
{
  Logger *log ;
  LoggerGpc *logR ;
    
  log = (Logger *) data->user_data ;
  
  logR = (LoggerGpc *) log->internal ;
  
  switch (arg)
  {
    case 0: /* Gpc since reset */
      *result = logR->r ;
      return LOGGER_OK ;
    break ;
      
    case 1: /* Reset date / time */
      *result = (double) logR->t ;
      return LOGGER_OK ;
    break ;
      
    case 2: /* Gpc rate */
      *result = logR->r_sum / (double) log->N ;
      return LOGGER_OK ;
    break ;
     
    case 3: /* Monotonic count */
      *result = (double) logR->c ;
      return LOGGER_OK ;
    break ;
     
    case 4: /* Delta count */
      *result = (double) logR->d ;
      return LOGGER_OK ;
    break ;
  }
  
  return LOGGER_ERROR ;    
}

int
logger_reset(owwl_conn *conn, owwl_data *data, void *user_data)
{  
  Logger *log ;
  
  log = (Logger *) data->user_data ;
  
  if (log->reset != NULL) log->reset(data, user_data) ;
  
  log->N = 0 ;
  
  return LOGGER_OK ;
}

int
logger_accumulate(owwl_conn *conn, owwl_data *data, void *user_data)
{  
  Logger *log ;
  
  log = (Logger *) data->user_data ;
  
  if (log->accumulate != NULL) log->accumulate(data, user_data) ;
  
  return LOGGER_OK ;
}

int
logger_calculate(owwl_conn *conn, owwl_data *data, void *user_data)
{  
  Logger *log ;
  
  log = (Logger *) data->user_data ;
  
  if (log->calculate != NULL) return log->calculate(data, user_data) ;
  
  return LOGGER_OK ;
}

int
logger_result(owwl_data *data, void *user_data, double *result, int arg)
{  
  Logger *log ;
  
  log = (Logger *) data->user_data ;
  
  if (
       (log->result != NULL) && 
       (log->result(data, user_data, result, arg) == LOGGER_OK)
     )
    return LOGGER_OK ;
    
  return LOGGER_ERROR ;
}

static int
logger_rain_result_string(owwl_data *data, void *user_data, int arg, char *buffer, int bufflen, char *format)
{
  double result ;
  Logger *log ;
  LoggerRain *logR ;
    
  log = (Logger *) data->user_data ;
  
  logR = (LoggerRain *) log->internal ;
  
  switch (arg)
  {
    case 0: /* Rain since reset */
    case 2: /* Rain rate */
      if (logger_result(data, user_data, &result, arg) == LOGGER_OK)
      {
        snprintf(buffer, bufflen, format, result) ;
        return LOGGER_OK ;
      }
      break ;
      
    case 1: /* Reset date / time */
      strftime(buffer, bufflen, "%Y-%m-%d %H:%M:%S", localtime(&logR->t)) ;
      return LOGGER_OK ;
      break ;
      
    case 3: /* Monotonic count */
      snprintf(buffer, bufflen, "%lu", logR->c) ;
      return LOGGER_OK ;
    break ;
  }
  
  buffer[0] = '\0' ;
  return LOGGER_ERROR ;
}

static int
logger_gpc_result_string(owwl_data *data, void *user_data, int arg, char *buffer, int bufflen, char *format)
{
  double result ;
  Logger *log ;
  LoggerGpc *logG ;
    
  log = (Logger *) data->user_data ;
  
  logG = (LoggerGpc *) log->internal ;
  
  switch (arg)
  {
    case 0: /* GPC since reset */
    case 2: /* Gpc rate */
      if (logger_result(data, user_data, &result, arg) == LOGGER_OK)
      {
        snprintf(buffer, bufflen, format, result) ;
        return LOGGER_OK ;
      }
      break ;
      
    case 1: /* Reset date / time */
      strftime(buffer, bufflen, "%Y-%m-%d %H:%M:%S", localtime(&logG->t)) ;
      return LOGGER_OK ;
      break ;
      
    case 3: /* Monotonic count */
      snprintf(buffer, bufflen, "%lu", logG->c) ;
      return LOGGER_OK ;
      break ;
    
    case 4: /* Delta */
      snprintf(buffer, bufflen, "%lu", logG->d) ;
      return LOGGER_OK ;
      break ;
  }
  
  buffer[0] = '\0' ;
  return LOGGER_ERROR ;
}

int
logger_result_string(owwl_data *data, void *user_data, int arg, char *buffer, int bufflen, char *format)
{
  double result ;
  
  switch  (data->device_type)
  {
    case OwwlDev_Rain:
      return logger_rain_result_string(data, user_data, arg, buffer, bufflen, format) ;
      
    case OwwlDev_Gpc:
      return logger_gpc_result_string(data, user_data, arg, buffer, bufflen, format) ;
      
    default:

      if (logger_result(data, user_data, &result, arg) == LOGGER_OK)
      {
        snprintf(buffer, bufflen, format, result) ;
        return LOGGER_OK ;
      }
  }
  
  buffer[0] = '\0' ;
  return LOGGER_ERROR ;
}

static void
logger_built_floats(Logger *log, int size)
{
  LoggerFloats *floats ;
  
  floats = (LoggerFloats *) calloc(1, sizeof(LoggerFloats)) ;
  floats->sum = (double *) calloc(size, sizeof(double)) ;
  log->internal = (void *) floats ;
  
  floats->count = log->results = size ;
  
  log->reset = logger_reset_floats ;
  log->accumulate = logger_accumulate_floats ;
  log->calculate = NULL ;
  log->result = logger_result_floats ;
}

static void
logger_build_wind(Logger *log, int size)
{
  LoggerWind *wind ;
  
  wind = (LoggerWind *) calloc(1, sizeof(LoggerWind)) ;
  wind->s_sum = wind->x_sum = wind->y_sum = wind->s_max = 0.0 ;
  log->internal = (void *) wind ;
  
  log->reset = logger_reset_wind ;
  log->accumulate = logger_accumulate_wind ;
  log->calculate = NULL ;
  log->result = logger_result_wind ;
  log->results = 3 ;
}

static void
logger_build_rain(Logger *log, int size)
{
  LoggerRain *rain ;
  
  rain = (LoggerRain *) calloc(1, sizeof(LoggerRain)) ;
  rain->c = 0L ;
  rain->r_sum = rain->r = 0.0 ;
  rain->t = 0 ;
  
  log->internal = (void *) rain ;
  
  log->reset = logger_reset_rain ;
  log->accumulate = logger_accumulate_rain ;
  log->calculate = NULL ;
  log->result = logger_result_rain ;
  log->results = 4 ;
}

static void
logger_build_gpc(Logger *log, int size)
{
  LoggerGpc *gpc ;
  
  gpc = (LoggerGpc *) calloc(1, sizeof(LoggerGpc)) ;
  gpc->c = gpc->d = 0L ;
  gpc->r_sum = gpc->r = 0.0 ;
  gpc->t = 0 ;
  
  log->internal = (void *) gpc ;
  
  log->reset = logger_reset_gpc ;
  log->accumulate = logger_accumulate_gpc ;
  log->calculate = NULL ;
  log->result = logger_result_gpc ;
  log->results = 5 ;
}


/* Construct logger struct for a data stream */
void
logger_build_data(owwl_conn *conn, owwl_data *data, void *user_data)
{
  Logger *log ;
  
  log = (Logger *) calloc(1, sizeof(Logger)) ;
  
  if (!log)
    owwlog_error("logger_build_data - memory allocation failed\n") ;
    
  data->user_data = (void *) log ;
    
  switch (data->device_type)
  {
    case OwwlDev_Temperature:
    case OwwlDev_Cryo:
    case OwwlDev_Diode:
    case OwwlDev_Barometer:
    case OwwlDev_Humidity:
    case OwwlDev_Solar:
    case OwwlDev_Uvi:
    case OwwlDev_Resistance:
    case OwwlDev_Heater:
    {
      logger_built_floats(log, 1) ;
      break ;
    }
    
    case OwwlDev_Psb:
    {
      logger_built_floats(log, 2) ;
      break ;
    }
      
    case OwwlDev_Grt:
    case OwwlDev_Adc:
    {
      logger_built_floats(log, 3) ;
      break ;
    }
    
    case OwwlDev_Wind:
    {
      logger_build_wind(log, 3) ;
      break ;
    }
   
    case OwwlDev_Rain:
    {
      logger_build_rain(log, 3) ;
      break ;
    }
       
    case OwwlDev_Gpc:
    {
      logger_build_gpc(log, 3) ;
      break ;
    }
       
    default: /* Do nothing */
      break ;
  }
}
